@@ -20,56 +20,51 @@
 #include <charconv>
 #include <memory>
 
-#include "driver/common/utils.h"
+#define ADBC_FRAMEWORK_USE_FMT
+#include "driver/framework/status.h"
 #include "error.h"
 
 namespace adbcpq {
 
 PqResultHelper::~PqResultHelper() { ClearResult(); }
 
-AdbcStatusCode PqResultHelper::PrepareInternal(int n_params, const Oid* param_oids,
-                                               struct AdbcError* error) {
+Status PqResultHelper::PrepareInternal(int n_params, const Oid* param_oids) const {
   // TODO: make stmtName a unique identifier?
   PGresult* result =
       PQprepare(conn_, /*stmtName=*/"", query_.c_str(), n_params, param_oids);
   if (PQresultStatus(result) != PGRES_COMMAND_OK) {
-    AdbcStatusCode code =
-        SetError(error, result, "[libpq] Failed to prepare query: %s\nQuery was:%s",
-                 PQerrorMessage(conn_), query_.c_str());
+    auto status = MakeStatus(result, "Failed to prepare query: {}\nQuery was:{}",
+                             PQerrorMessage(conn_), query_.c_str());
     PQclear(result);
-    return code;
+    return status;
   }
 
   PQclear(result);
-  return ADBC_STATUS_OK;
+  return Status::Ok();
 }
 
-AdbcStatusCode PqResultHelper::Prepare(struct AdbcError* error) {
-  return PrepareInternal(0, nullptr, error);
+Status PqResultHelper::Prepare() const { return PrepareInternal(0, nullptr); }
+
+Status PqResultHelper::Prepare(const std::vector<Oid>& param_oids) const {
+  return PrepareInternal(static_cast<int>(param_oids.size()), param_oids.data());
 }
 
-AdbcStatusCode PqResultHelper::Prepare(const std::vector<Oid>& param_oids,
-                                       struct AdbcError* error) {
-  return PrepareInternal(param_oids.size(), param_oids.data(), error);
-}
-
-AdbcStatusCode PqResultHelper::DescribePrepared(struct AdbcError* error) {
+Status PqResultHelper::DescribePrepared() {
   ClearResult();
   result_ = PQdescribePrepared(conn_, /*stmtName=*/"");
   if (PQresultStatus(result_) != PGRES_COMMAND_OK) {
-    AdbcStatusCode code = SetError(
-        error, result_, "[libpq] Failed to describe prepared statement: %s\nQuery was:%s",
+    Status status = MakeStatus(
+        result_, "[libpq] Failed to describe prepared statement: {}\nQuery was:{}",
         PQerrorMessage(conn_), query_.c_str());
     ClearResult();
-    return code;
+    return status;
   }
 
-  return ADBC_STATUS_OK;
+  return Status::Ok();
 }
 
-AdbcStatusCode PqResultHelper::Execute(struct AdbcError* error,
-                                       const std::vector<std::string>& params,
-                                       PostgresType* param_types) {
+Status PqResultHelper::Execute(const std::vector<std::string>& params,
+                               PostgresType* param_types) {
   if (params.size() == 0 && param_types == nullptr && output_format_ == Format::kText) {
     ClearResult();
     result_ = PQexec(conn_, query_.c_str());
@@ -95,23 +90,21 @@ AdbcStatusCode PqResultHelper::Execute(struct AdbcError* error,
     }
 
     ClearResult();
-    result_ = PQexecParams(conn_, query_.c_str(), param_values.size(), param_oids_ptr,
-                           param_values.data(), param_lengths.data(),
+    result_ = PQexecParams(conn_, query_.c_str(), static_cast<int>(param_values.size()),
+                           param_oids_ptr, param_values.data(), param_lengths.data(),
                            param_formats.data(), static_cast<int>(output_format_));
   }
 
   ExecStatusType status = PQresultStatus(result_);
   if (status != PGRES_TUPLES_OK && status != PGRES_COMMAND_OK) {
-    AdbcStatusCode status =
-        SetError(error, result_, "[libpq] Failed to execute query '%s': %s",
-                 query_.c_str(), PQerrorMessage(conn_));
-    return status;
+    return MakeStatus(result_, "[libpq] Failed to execute query '{}': {}", query_.c_str(),
+                      PQerrorMessage(conn_));
   }
 
-  return ADBC_STATUS_OK;
+  return Status::Ok();
 }
 
-AdbcStatusCode PqResultHelper::ExecuteCopy(struct AdbcError* error) {
+Status PqResultHelper::ExecuteCopy() {
   // Remove trailing semicolon(s) from the query before feeding it into COPY
   while (!query_.empty() && query_.back() == ';') {
     query_.pop_back();
@@ -125,20 +118,19 @@ AdbcStatusCode PqResultHelper::ExecuteCopy(struct AdbcError* error) {
                          static_cast<int>(Format::kBinary));
 
   if (PQresultStatus(result_) != PGRES_COPY_OUT) {
-    AdbcStatusCode code = SetError(
-        error, result_,
-        "[libpq] Failed to execute query: could not begin COPY: %s\nQuery was: %s",
+    Status status = MakeStatus(
+        result_,
+        "[libpq] Failed to execute query: could not begin COPY: {}\nQuery was: {}",
         PQerrorMessage(conn_), copy_query.c_str());
     ClearResult();
-    return code;
+    return status;
   }
 
-  return ADBC_STATUS_OK;
+  return Status::Ok();
 }
 
-AdbcStatusCode PqResultHelper::ResolveParamTypes(PostgresTypeResolver& type_resolver,
-                                                 PostgresType* param_types,
-                                                 struct AdbcError* error) {
+Status PqResultHelper::ResolveParamTypes(PostgresTypeResolver& type_resolver,
+                                         PostgresType* param_types) {
   struct ArrowError na_error;
   ArrowErrorInit(&na_error);
 
@@ -149,22 +141,22 @@ AdbcStatusCode PqResultHelper::ResolveParamTypes(PostgresTypeResolver& type_reso
     const Oid pg_oid = PQparamtype(result_, i);
     PostgresType pg_type;
     if (type_resolver.Find(pg_oid, &pg_type, &na_error) != NANOARROW_OK) {
-      SetError(error, "%s%d%s%s%s%d", "[libpq] Parameter #", i + 1, " (\"",
-               PQfname(result_, i), "\") has unknown type code ", pg_oid);
+      Status status = Status::NotImplemented("[libpq] Parameter #", i + 1, " (\"",
+                                             PQfname(result_, i),
+                                             "\") has unknown type code ", pg_oid);
       ClearResult();
-      return ADBC_STATUS_NOT_IMPLEMENTED;
+      return status;
     }
 
     root_type.AppendChild(PQfname(result_, i), pg_type);
   }
 
   *param_types = root_type;
-  return ADBC_STATUS_OK;
+  return Status::Ok();
 }
 
-AdbcStatusCode PqResultHelper::ResolveOutputTypes(PostgresTypeResolver& type_resolver,
-                                                  PostgresType* result_types,
-                                                  struct AdbcError* error) {
+Status PqResultHelper::ResolveOutputTypes(PostgresTypeResolver& type_resolver,
+                                          PostgresType* result_types) {
   struct ArrowError na_error;
   ArrowErrorInit(&na_error);
 
@@ -175,17 +167,17 @@ AdbcStatusCode PqResultHelper::ResolveOutputTypes(PostgresTypeResolver& type_res
     const Oid pg_oid = PQftype(result_, i);
     PostgresType pg_type;
     if (type_resolver.Find(pg_oid, &pg_type, &na_error) != NANOARROW_OK) {
-      SetError(error, "%s%d%s%s%s%d", "[libpq] Column #", i + 1, " (\"",
-               PQfname(result_, i), "\") has unknown type code ", pg_oid);
-      ClearResult();
-      return ADBC_STATUS_NOT_IMPLEMENTED;
+      // We couldn't look up the OID.
+      // TODO(apache/arrow-adbc#1243): issue a warning (maybe reloading the
+      // connection will load the OIDs if it was a newly created type)
+      pg_type = PostgresType::Unnamed(pg_oid);
     }
 
     root_type.AppendChild(PQfname(result_, i), pg_type);
   }
 
   *result_types = root_type;
-  return ADBC_STATUS_OK;
+  return Status::Ok();
 }
 
 PGresult* PqResultHelper::ReleaseResult() {
@@ -194,7 +186,7 @@ PGresult* PqResultHelper::ReleaseResult() {
   return out;
 }
 
-int64_t PqResultHelper::AffectedRows() {
+int64_t PqResultHelper::AffectedRows() const {
   if (result_ == nullptr) {
     return -1;
   }
